%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% DBA.m
%
% An implementation of Dynamic Time Warping Barycenter Averaging.
%
% François Petitjean, Alain Ketterlin, and Pierre Gançarski. 2011.
% A global averaging method for dynamic time warping, with applications to clustering.
% Pattern Recogn. 44, 3 (March 2011), 678-693.
% DOI=10.1016/j.patcog.2010.09.013 http://dx.doi.org/10.1016/j.patcog.2010.09.013
%
function [C, CI, SI] = DBA(sequences, maxiter, C, distance)

  if nargin < 2
    maxiter = 100;
  end

  if nargin < 3
    C = sequences{randi(length(sequences))};
  end

  if nargin < 4
    distance = @EuclidDistance;
  end

  CI = {};
  SI = {};

  for i = 1:maxiter
    for j = 1:length(sequences)
      [C, CI{j}, SI{j}] = DBAStep(C, sequences{j}, distance);
    end
  end

end

function [C, CI, SI] = DBAStep(C, S, distance)

  D = DTWDistance(C, S, distance);
  [CI, SI] = DTWPath(D);

  for i = 1:size(C, 1)
    C(i, :) = mean(S(SI(find(CI == i))), 1);
  end

end

function D = DTWDistance(A, B, distance)

  D = zeros(size(A, 1), size(B, 1));

  D(1, 1) = distance(A(1, :), B(1, :));

  for ai = 2:size(A, 1)
    D(ai, 1) = D(ai - 1, 1) + distance(A(ai, :), B(1, :));
  end

  for bi = 2:size(B, 1)
    D(1, bi) = D(1, bi - 1) + distance(A(1, :), B(bi, :));
  end

  for ai = 2:size(A, 1)
    for bi = 2:size(B, 1)
      D(ai, bi) = min([D(ai - 1, bi); D(ai - 1, bi - 1); D(ai, bi - 1)]) + distance(A(ai, :), B(bi, :));
    end
  end

end

function [AI, BI] = DTWPath(D)

  AI = [];
  BI = [];

  ai = size(D, 1);
  bi = size(D, 2);

  while ai > 1 && bi > 1
    AI(end + 1) = ai;
    BI(end + 1) = bi;
    cands = [D(ai - 1, bi); D(ai - 1, bi - 1); D(ai, bi - 1)];
    idx = min(find(cands == min(cands)));
    if idx < 3
      ai = ai - 1;
    end
    if idx > 1
      bi = bi - 1;
    end
  end

  if ai == 1
    AI(end+1:end+bi) = ones(bi, 1);
    BI(end+1:end+bi) = linspace(bi, 1, bi);
  else
    AI(end+1:end+ai) = linspace(ai, 1, ai);
    BI(end+1:end+ai) = ones(ai, 1);
  end

  AI = flipud(AI');
  BI = flipud(BI');

end

function distance = EuclidDistance(a, b)
  distance = sqrt(sum((a - b) .^ 2));
end
