%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% GaussianProcessModelRotate.m
%
classdef GaussianProcessModelRotate < Model

  properties (Access = private)
    algorithm
    kernelGeneratorStrategy

    currentKernels
    prevKernels
    prevGramMatrix
    prevSurface
  end

  methods (Access = public)

    function this = GaussianProcessModelRotate(kind, algorithm)
      this@Model(kind);
      this.algorithm = algorithm;
      kernelGeneratorStrategyHandle = ...
          algorithm.GetKernelGeneratorStrategyHandle();
      this.kernelGeneratorStrategy = ...
          kernelGeneratorStrategyHandle(algorithm, kind);
    end

    function Initialize(this, mcdcInput, mcdcOutput, aspect)
      this.kernelGeneratorStrategy.SetInitialKernelGens( ...
          this.Kind.GetKernelGenerators(mcdcInput));
    end

    function f = GetFunction(this, j, in, out, aspect, gridData)
      aspect.Log('  Sampling %s model from GP', char(this.Kind));

      % State 0: Prepare grid points on which gram matrix is calculated
      gramMatrixGrid = this.algorithm.getGramMatrixGrid(in, gridData, aspect);
      % Stage 1: Sample kernel parameters
      kernels = this.GenerateKernel(j, in, out, aspect);
      this.currentKernels = kernels;
      % Stage 3: Construct gram matrix on grid points
      gramMatrix = this.CreateGramMatrix(j, kernels, gramMatrixGrid, aspect);
      this.gramMatrix = gramMatrix;
      % Stage 4: Draw realization of the GP surface
      eqnVals = this.DrawGPSurface(gridData, gramMatrix, j, in, out, aspect);
      this.eqnVals = eqnVals;
      % Stage 5: Construct a B3 spline approximation
      splineHandle = this.Kind.GetSplineHandle(in);
      f = this.Spline(splineHandle, gridData, eqnVals);
    end

    function f = GetFunctionFromCurrentParams(this, j, in, out, aspect, gridData)
      % Stage 4: Draw realization of the GP surface
      eqnVals = this.DrawGPSurface(gridData, this.gramMatrix, j, in, out, aspect);
      this.eqnVals = eqnVals;
      % Stage 5: Construct a B3 spline approximation
      splineHandle = this.Kind.GetSplineHandle(in);
      f = this.Spline(splineHandle, gridData, eqnVals);
    end

    function OnAccept(this, j, in, out, aspect)
      this.Kind.AddKernelParameters(out, this.GetCurrentKernelParameters());
      this.Kind.AddEquationValues(out, this.eqnVals);
      this.prevKernels = this.currentKernels;
      this.prevGramMatrix = this.gramMatrix;
      this.prevSurface = this.eqnVals;
    end

  end

  methods (Access = private)

    function kernels = GenerateKernel(this, j, in, out, aspect)
      kernelGenerator = this.kernelGeneratorStrategy.GetGenerator(j, in, out);
      dims = length(kernelGenerator);
      for i = 1:dims
        if (j == 1 || mod(j, dims) == i - 1)
          kernels{i} = kernelGenerator{i}.Generate();
          aspect.Log('  %s[%d]: %s', char(this.Kind), i, kernels{i}.ToString());
        else
          kernels{i} = this.prevKernels{i};
        end
      end
    end

    function gramMatrix = CreateGramMatrix(this, j, kernels, gramMatrixGrid, aspect)
      dims = length(kernels);
      for i = 1:dims
        if (j == 1 || mod(j, dims) == i - 1)
          gramMatrix{i} = kernels{i}.CreateGramMatrix(gramMatrixGrid);
          aspect.Log('    %sKernel[%d] Done', char(this.Kind), i);
        else
          gramMatrix{i} = this.prevGramMatrix{i};
        end
      end
    end

    function surface = DrawGPSurface(this, gridData, gramMatrix, j, in, out, aspect)
      aspect.Log('  Drawing GP surface for %s...', char(this.Kind));
      mcdcStrategy = this.algorithm.GetStrategy();
      meanFunction = this.Kind.GetMeanFunction(mcdcStrategy);
      covarianceFunction = this.Kind.GetCovarianceFunction(mcdcStrategy);
      surface = this.algorithm.DrawGPSurface( ...
        gridData, ...
        in.gridDimForGramMatrix, ...
        meanFunction(j, in, out), ...
        covarianceFunction(j, gramMatrix));
      dims = length(gramMatrix);
      for i = 1:dims
        if (j > 1 && mod(j, dims) ~= i - 1)
          surface(i, :) = this.prevSurface(i, :);
        end
      end
    end

    function modelFunction = Spline(this, splineHandle, gridData, surface)
      modelFunction = VectorValuedFunction(ModelFunctions2(splineHandle, gridData, surface));
    end

    function kernelParameters = GetCurrentKernelParameters(this)
      kernelParameters = cell2mat(arrayfun( ...
          @(i) this.currentKernels{i}.Properties(), ...
          1:length(this.currentKernels), ...
          'UniformOutput', ...
          false))';
    end

  end

end
