%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% Graphs.m
%
classdef Graphs < handle

  methods (Static)

    %
    % plot values of observation variable Y through the iterations
    % The graphs will be generated for each dimension of Y.
    %
    % YE(outputFileNamePrefix, matFileName, iterations)
    % YE(outputFileNamePrefix, matFileName, iterations, times)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: matfile of data to be plotted
    %   iterations: Array of iterations for plot
    %   times: range of time for graphs. When omitted, graphs will be generated through all time
    %
    function YE(outputFileNamePrefix, matFileName, iterations, varargin)
      load(matFileName);
      % in.z, out.j, out.YE loaded

      J = out.j;
      T = size(out.YE, 2);
      dim = size(out.YE, 3);
      if (nargin >= 4)
        orgTs = varargin{1};
        ts = orgTs(1 <= orgTs & orgTs <= T);
      else
        ts = 1:T;
      end

      for d = 1:dim
        g = PlotGraph();
        g.Title('Estimated value of observation through the iteration');
        g.XLabel('Time');
        %        g.YLabel('Y');
        % z
        g.AddSeries('Obseravtion', ts, in.z(d, ts));
        % YE
        for i = iterations
          if (1 <= i && i <= J)
            ye = out.YEAt(i);
            g.AddSeries(num2str(i), ts, ye(1, ts, d));
          end
        end

        if (dim == 1)
          g.SaveAs(Graphs.pdfFileName(outputFileNamePrefix));
        else
          g.SaveAs(Graphs.pdfNumberedFileName(outputFileNamePrefix, d));
        end
      end
    end

    %
    % plot values of state variable X through the iterations
    % The graphs will be generated for each dimension of X.
    %
    % XE(outputFileNamePrefix, matFileName, iterations)
    % XE(outputFileNamePrefix, matFileName, iterations, times)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: matfile of data to be plotted
    %   iterations: Array of iterations for plot
    %   times: range of time for graphs. When omitted, graphs will be generated through all time
    %
    function XE(outputFileNamePrefix, matFileName, iterations, varargin)
      load(matFileName);
      % out.j, out.XE loaded.

      J = out.j;
      T = size(out.XE, 2);
      dim = size(out.XE, 3);
      if (nargin >= 4)
        orgTs = varargin{1};
        ts = orgTs(1 <= orgTs & orgTs <= T);
      else
        ts = 1:T;
      end

      for d = 1:dim
        g = PlotGraph();
        g.Title('Estimated value of state through the iteration');
        g.XLabel('Time');
        g.YLabel('State');
        for i = iterations
          if (1 <= i && i <= J)
            xe = out.XEAt(i);
            g.AddSeries(num2str(i), ts, xe(1, ts, d));
          end
        end

        if (dim == 1)
          g.SaveAs(Graphs.pdfFileName(outputFileNamePrefix));
        else
          g.SaveAs(Graphs.pdfNumberedFileName(outputFileNamePrefix, d));
        end
      end
    end

    %
    % plot average etimated values of obserbation variable Y through the iterations
    % The graphs will be generated for each dimension of Y.
    %
    % YEMean(outputFileNamePrefix, matFileName)
    % YEMean(outputFileNamePrefix, matFileName, times)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: matfile of data to be plotted
    %   times: range of time for graphs. When omitted, graphs will be generated through all time
    %
    function YEMean(outputFileNamePrefix, matFileName, varargin)
      load(matFileName);
      % in.z, out.j, out.YE loaded.

      J = out.j;
      T = size(out.YE, 2);
      dim = size(out.YE, 3);
      if (nargin >= 3)
        orgTs = varargin{1};
        ts = orgTs(1 <= orgTs & orgTs <= T);
      else
        ts = 1:T;
      end

      YE = out.YEAt(1:J);

      if (dim == 1)
        YEMean = mean(YE(:, ts));
        YEStd = std(YE(:, ts));
      else
        YEMean = squeeze(mean(YE(:, ts, :), 1));
        YEStd = squeeze(std(YE(:, ts, :), 1));
      end

      for d = 1:dim
        g = PlotGraph();
        g.Title('Average estimated values of obserbation through the iterations');
        g.XLabel('Time');
        % g.YLabel('Y');
        % z
        g.AddSeries('Observation', ts, in.z(d, ts), 'Color', 'red');

        if (dim == 1)
          g.AddSeries('Mean', ts, YEMean, 'Color', 'blue');
          g.AddSeries('+\sigma (standard dev.)', ts, YEMean + YEStd, 'Color', 'green', 'LineStyle', '--');
          g.AddSeries('-\sigma (standard dev.)', ts, YEMean - YEStd, 'Color', 'green', 'LineStyle', '--');
          g.SaveAs(Graphs.pdfFileName(outputFileNamePrefix));
        else
          g.AddSeries('Mean', ts, YEMean(:, d), 'Color', 'blue');
          g.AddSeries('+\sigma (standard dev.)', ts, YEMean(:, d) + YEStd(:, d), 'Color', 'green', 'LineStyle', '--');
          g.AddSeries('-\sigma (standard dev.)', ts, YEMean(:, d) - YEStd(:, d), 'Color', 'green', 'LineStyle', '--');
          g.SaveAs(Graphs.pdfNumberedFileName(outputFileNamePrefix, d));
        end
      end
    end

    %
    % plot average etimated values of state variable X through the iterations
    % The graphs will be generated for each dimension of X
    %
    % XEMean(outputFileNamePrefix, x, matFileName)
    % XEMean(outputFileNamePrefix, x, matFileName, times)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: matfile of data to be plotted
    %   times: range of time for graphs. When omitted, graphs will be generated through all time
    %
    function XEMean(outputFileNamePrefix, matFileName, varargin)
      load(matFileName);
      % out.j, out.XE loaded.

      J = out.j;
      T = size(out.XE, 2);
      dim = size(out.XE, 3);
      if (nargin >= 3)
        orgTs = varargin{1};
        ts = orgTs(1 <= orgTs & orgTs <= T);
      else
        ts = 1:T;
      end

      XE = out.XEAt(1:J);

      if (dim == 1)
        XEMean = mean(XE(:, ts));
        XEStd = std(XE(:, ts));
      else
        XEMean = squeeze(mean(XE(:, ts, :), 1));
        XEStd = squeeze(std(XE(:, ts, :), 1));
      end

      for d = 1:dim
        g = PlotGraph();
        g.Title('Average estimated values of state through the iterations');
        g.XLabel('Time');
        g.YLabel('X');

        if (dim == 1)
          g.AddSeries('Mean', ts, XEMean, 'Color', 'blue');
          g.AddSeries('+\sigma (std. dev.)', ts, XEMean + XEStd, 'Color', 'green', 'LineStyle', '--');
          g.AddSeries('-\sigma (std. dev.)', ts, XEMean - XEStd, 'Color', 'green', 'LineStyle', '--');
          g.SaveAs(Graphs.pdfFileName(outputFileNamePrefix));
        else
          g.AddSeries('Mean', ts, XEMean(:, d), 'Color', 'blue');
          g.AddSeries('+\sigma (std. dev.)', ts, XEMean(:, d) + XEStd(:, d), 'Color', 'green', 'LineStyle', '--');
          g.AddSeries('-\sigma (std. dev.)', ts, XEMean(:, d) - XEStd(:, d), 'Color', 'green', 'LineStyle', '--');
          g.SaveAs(Graphs.pdfNumberedFileName(outputFileNamePrefix, d));
        end
      end
    end

    %
    % plot graph of log-likelihood through the iterations
    %
    % Loglik = Loglik(outputFileName, matFileName1, matFileName2, ...)
    % input
    %   outputFileName: output graph file name (PDF)
    %   matFileName1, matFileName2, ...: .mat file (multiple FILES can be specifed.)
    % output
    %   Loglik: log-likelihood
    %
    function Loglik = Loglik(outputFileName, varargin)
      Loglik = Graphs.CalcLoglik(varargin{:});

      g = PlotGraph();
      g.Title('Log likelihood');
      g.XLabel('Iteration');
      g.YLabel('Log likelihood');
      for i = 1:length(Loglik)
        matFileName = varargin{i};
        x = 1:length(Loglik{i});
        y = Loglik{i};
        g.AddSeries(matFileName, x, y);
      end
      g.SaveAs(Graphs.pdfFileName(outputFileName));
    end

    %
    % calculate log-likelihood through the iterations
    %
    % Loglik = CalcLoglik(matFileName1, matFileName2, ...)
    % input
    %   matFileName1, matFileName2, ...: .mat file (multiple files can be specifed.)
    % output
    %   Loglik: log-likelihood
    %
    function Loglik = CalcLoglik(varargin)
      if (nargin < 1)
        error('Specify matFileName(s).');
      end

      Loglik = {};
      for mfn = varargin
        matFileName = mfn{1};
        load(matFileName);
        % out.j, out.loglik loaded.
        Loglik{length(Loglik) + 1} = out.loglikAt(1:out.j)';
      end
    end

    % plot root mean squared error through the iterations
    %
    % [RmseMean, RmseStd, Rmse] = Rmse(outputFileName, matFileName)
    % input
    %   outputFileName: output file name (PDF)
    %   matFileName: matfile of data to be plotted
    % output
    %   RmseMean: Mean of RMSE
    %   RmseStd:  Standard deviation of RMSE
    %   Rmse:     RMSE
    %
    function [RmseMean, RmseStd, Rmse] = Rmse(outputFileName, matFileName)
      [RmseMean, RmseStd, Rmse] = Graphs.CalcRmse(matFileName);

      J = length(RmseMean);

      g = PlotGraph();
      g.Title('RMSE');
      g.XLabel('Iteration');
      g.YLabel('RMSE');
      g.AddSeries('Mean', 1:J, RmseMean, 'Color', 'blue');
      g.AddSeries('+\sigma (std. dev.)', 1:J, RmseMean + RmseStd, 'Color', 'green', 'LineStyle', '--');
      g.AddSeries('-\sigma (std. dev.)', 1:J, RmseMean - RmseStd, 'Color', 'green', 'LineStyle', '--');
      g.SaveAs(Graphs.pdfFileName(outputFileName));
    end

    %
    % Calculated root mean squared error through the iterations
    %
    % [RmseMean, RmseStd, Rmse] = CalcRmse(matFileName)
    %   matFileName: input filename .mat
    % Return value
    %   RmseMean: mean of RMSE
    %   RmseStd:  standard deviation of RMSE
    %   Rmse:     RMSE
    %
    function [RmseMean, RmseStd, Rmse] = CalcRmse(matFileName)
      load(matFileName);
      % in.z, out.j, out.YE loaded.

      J = out.j;
      T = size(out.YE, 2);
      dim = size(out.YE, 3);

      YE_mean = cumsum(out.YEAt(1:J), 1) ./ repmat([1:J]', 1, T, dim);
      Rmse = sqrt(sum((YE_mean - permute(repmat(in.z', 1, 1, J), [3 1 2])) .^ 2, 3));
      RmseMean = mean(Rmse, 2);
      RmseStd = std(Rmse, 0, 2);
    end

    %
    % Plot and generate the graphs of the estimated state transition function 'f'.
    % When the state is multidimensional, the values will be fixed except the one dimension where the value is varied.
    % The graphs will be generated for each dimension of observation.
    %
    % FV(outputFileNamePrefix, matFileName, iterations, x1, x2, x3, ...)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: Filename for output data matfile
    %   iterations: Array of iterations for plot
    %   x1, x2, x3, ...: fixed value or range for each dim.
    %
    % example:
    % Generate graphs for state transition function at iteration 1, 10, 100
    % where the dimension of state is 3, the range for first dim. is [-1 : .1 : 1], the fixed value for second dim. is 0,
    % and the fixed value for third dim. is 0.
    % >> Graphs.FV('test', 'test.mat', [1 10 100], -1:.1:1, 0, 0)
    %
    function FV(outputFileNamePrefix, ...
                matFileName, ...
                iterations, ...
                varargin)
      load(matFileName);
      % in.stateSplineHandle, in.xGrids, out.j, out.FV are loaded beforehand.
      Graphs.fvOrGvGraph( ...
          'FV', ...
          'X(t+1)', ...
          outputFileNamePrefix, ...
          in.stateSplineHandle, ...
          iterations, ...
          in.xGrids, ...
          @(i) out.FVAt(i), ...
          out.j, ...
          varargin{:});
    end

    %
    % Plot and generate the graphs of the estimated observation function 'g'
    % When the state is multidimensional, the values will be fixed except the one dimension where the value is varied.
    % The graphs will be generated for each dimension of observation.
    %
    % GV(outputFileNamePrefix, matFileName, iterations, x1, x2, x3, ...)
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: Filename for output data matfile
    %   iterations: Array of iterations for plot
    %   x1, x2, x3, ...: fixed value or range for each dim.
    %
    % Example:
    % Generate graphs for obserbation function at iteration 1, 10, 100
    % where the dimension of state is 3, the range for first dim. is [-1 : .1 : 1], the fixed value for second dim. is 0,
    % and the fixed value for third dim. is 0.
    % >> Graphs.GV('test', 'test.mat', [1 10 100], -1:.1:1, 0, 0)
    %
    function GV(outputFileNamePrefix, ...
                matFileName, ...
                iterations, ...
                varargin)
      load(matFileName);
      % in.obsSplineHandle, in.xGrids, out.j, out.GV loaded.
      Graphs.fvOrGvGraph( ...
          'GV', ...
          'Y(t)', ...
          outputFileNamePrefix, ...
          in.obsSplineHandle, ...
          iterations, ...
          in.xGrids, ...
          @(i) out.GVAt(i), ...
          out.j, ...
          varargin{:});
    end

    % Plot graphs for each kernel hyper-parameter of state through the iterations
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: Filename for output data matfile
    %
    % SKP : State Kernel Parameter
    function SKP(outputFileNamePrefix, matFileName)
      load(matFileName);
      % out.j, out.SKP loaded.
      Graphs.kpGraph('SKP', outputFileNamePrefix, out.SKPAt(1:out.j));
    end

    % Plot graphs for each kernel hyper-parameter of state through the iterations
    %   outputFileNamePrefix: Prefix for output graphe file name (PDF)
    %   matFileName: Filename for output data matfile
    %
    % OKP : Obserbation Kernel Parameter
    function OKP(outputFileNamePrefix, matFileName)
      load(matFileName);
      % out.j, out.OKP loaded.
      Graphs.kpGraph('OKP', outputFileNamePrefix, out.OKPAt(1:out.j));
    end

  end

  methods (Static, Access = private)

    function result = pdfFileName(str)
      tokens = regexpi(str, '(.*)\.pdf$', 'tokens');
      if (size(tokens, 2) == 1)
        result = str;
      else
        result = sprintf('%s.pdf', str);
      end
    end

    function result = pdfNumberedFileName(str, i)
      tokens = regexpi(str, '(.*)\.pdf$', 'tokens');
      if (size(tokens, 2) == 1)
        result = sprintf('%s_%d.pdf', tokens{1}{1}, i);
      else
        result = sprintf('%s_%d.pdf', str, i);
      end
    end

    function fvOrGvGraph(title, ...
                         ylabel, ...
                         outputFileNamePrefix, ...
                         splineHandle, ...
                         iterations, ...
                         xGrids, ...
                         eqnValsAt, ...
                         J, ...
                         varargin)
      inDim = size(xGrids, 2);
      varargSize = nargin - 8;

      % varargin check
      if (varargSize ~= inDim)
        error(sprintf('Needs args for %d dimension(s).', inDim));
      end
      rangeDim = 0;
      rangeSize = 0;
      for i = 1:varargSize
        s = size(varargin{i}, 2);
        if (s > 1)
          if (rangeDim ~= 0)
            error('Not allowed to specify a range to multi dimensions.');
          end
          rangeDim = i;
          rangeSize = s;
        end
      end
      if (rangeDim == 0)
        error('Specify a range to a single dimension.');
      end
      graphX = varargin{rangeDim};
      funcX = [];
      for i = 1:varargSize
        if (i == rangeDim)
          funcX = vertcat(funcX, varargin{i});
        else
          funcX = vertcat(funcX, repmat(varargin{i}, 1, rangeSize));
        end
      end

      % function execute.
      validIterations = [];
      funcResults = [];
      gridData = GridData(xGrids);
      for i = iterations
        if (1 <= i && i <= J)
          validIterations = horzcat(validIterations, i);
          functions = ModelFunctions(splineHandle, gridData, squeeze(eqnValsAt(i)));
          f = VectorValuedFunction(functions);
          funcResults = cat(3, funcResults, f(funcX'));
        end
      end

      % Graph
      outDim = size(funcResults, 2);
      for d = 1:outDim
        g = PlotGraph();
        g.Title(title);
        g.XLabel('X(t)');
        g.YLabel(ylabel);
        for i = 1:size(validIterations, 2)
          g.AddSeries(num2str(validIterations(i)), graphX, funcResults(:, d, i)');
        end

        if (outDim == 1)
          g.SaveAs(Graphs.pdfFileName(outputFileNamePrefix));
        else
          g.SaveAs(Graphs.pdfNumberedFileName(outputFileNamePrefix, d));
        end
      end

    end

    function kpGraph(title, outputFileNamePrefix, kp)
      J = size(kp, 1);
      dim = size(kp, 2);
      pc = size(kp, 3);

      pnames = { 'Sigma', 'L' };
      for p = 1:2
        g = PlotGraph();
        g.Title(title);
        g.XLabel('Iteration');
        g.YLabel(pnames{p});
        for d = 1:dim
          g.AddSeries(sprintf('dim %d', d), 1:J, kp(1:J, d, p));
        end
        g.SaveAs(sprintf('%s_%s', outputFileNamePrefix, pnames{p}));
      end

    end

  end

end
