%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html
%
% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori
%
% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

classdef IterationStrategyDefault < IterationStrategy

  methods (Access = public)

    function this = IterationStrategyDefault()
    end

    function ProcessIteration(this, j, in, out, aspect, gridData)

      aspect.StartIteration();

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Sample model functions
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      fns{1} = this.stateModel.GetFunction(j, in, out, aspect, gridData);
      fno{1} = this.obsModel.GetFunction(j, in, out, aspect, gridData);
      for di = 2:this.numDraw
        fns{di} = this.stateModel.GetFunctionFromCurrentParams(j, in, out, aspect, gridData);
        fno{di} = this.obsModel.GetFunctionFromCurrentParams(j, in, out, aspect, gridData);
      end

      FnState = @(x) IterationStrategyDefault.meanGPFunction(fns, x);
      FnObs = @(x) IterationStrategyDefault.meanGPFunction(fno, x);

      GramMatState = this.stateModel.gramMatrix;
      GramMatObs = this.obsModel.gramMatrix;

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Stage 6: Estimate using particle filter
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      aspect.Log('  Estimating using particle filter... (# of particles = %d)', in.N);

      if isempty(this.stateNoise) || isempty(this.obsNoise)
        logLikelihood = this.likelihoodCalculator.LogLikelihood( ...
            FnState, FnObs);
      else
        logLikelihood = this.likelihoodCalculator.LogLikelihood( ...
            FnState, FnObs, this.stateNoise, this.obsNoise);
      end
      if (isa(this.likelihoodCalculator, 'ParticleFilter') && this.likelihoodCalculator.OGC(end) > 0)
        aspect.Log('  Total %d particles are out of the grid', this.likelihoodCalculator.OGC(end));
      end

      aspect.Log('  Log likelihood = %f', logLikelihood);

      acceptanceLogProbability = -Inf;
      accepted = false;
      try
        acceptanceLogProbability = this.mcdcStrategy.AcceptanceLogProbability( ...
            j, in, out, logLikelihood, this.stateModel, this.obsModel);
        accepted = this.isAccepted(j, acceptanceLogProbability);
      catch e
        aspect.Log('  !Error in calculation of acceptance probability');
        % Reject
      end

      if j == 1
        message = '';
      elseif accepted
        message = ', accepted';
      else
        message = ', rejected';
      end
      aspect.Log('  Acceptance log probability = %f%s', acceptanceLogProbability, message);

      if accepted
        this.stateModel.OnAccept(j, in, out, aspect);
        this.obsModel.OnAccept(j, in, out, aspect);
        this.likelihoodCalculator.OnAccept(j, in, out, aspect);
        out.AddFnState(FnState, j);
        out.AddFnObs(FnObs, j);
        this.mcdcStrategy.SaveGramMatrices(GramMatState, GramMatObs);
      end

      out.SetIDXAsLast(j);
      out.j = j;
      aspect.EndIteration();

    end

  end

  methods (Static, Access = protected)

    function result = meanGPFunction(funcs, x)
      n = length(funcs);
      result = funcs{1}(x);
      for i = 2:n
        result = result + funcs{i}(x);
      end
      result = result ./ n;
    end

  end

  methods (Access = private)

    function result = isAccepted(this, j, logProbability)
      result = (j == 1 || rand(1) < exp(logProbability));
    end

  end

end
