%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html
%
% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori
%
% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

classdef IterationStrategyMHGibbs < IterationStrategy

  properties (Access = private)
    thetaFPrev
    thetaGPrev
    fPrev
    gPrev
    xPrev
    yPrev
    u
    v
  end

  methods (Access = public)

    function ProcessIteration(this, j, in, out, aspect, gridData)

      aspect.StartIteration();

      this.thetaFPrev = this.BuildKernelFromParams(out.SKPAt(j - 1));
      this.thetaGPrev = this.BuildKernelFromParams(out.OKPAt(j - 1));
      this.fPrev = reshape(out.FVAt(j - 1), size(this.stateModel.eqnVals));
      this.gPrev = reshape(out.GVAt(j - 1), size(this.obsModel.eqnVals));
      this.xPrev = out.XEAt(j - 1);
      this.yPrev = out.YEAt(j - 1);

      changed = false;

      % Step 1
      thetaG = this.obsModel.MHGenerateKernel(j, in, out, aspect);
      prob1 = this.CalcStep1Prob(j, in, out, aspect, thetaG, gridData);
      if rand() < prob1
        changed = true;
        aspect.Log('  Step 1: accepted');
      else
        thetaG = this.thetaGPrev;
        this.obsModel.MHSetCurrentKernels(thetaG);
        aspect.Log('  Step 1: rejected');
      end
      gramMatG = this.obsModel.MHCreateGramMatrix(j, in, out, aspect, thetaG, gridData);

      % Step 2
      thetaF = this.stateModel.MHGenerateKernel(j, in, out, aspect);
      prob2 = this.CalcStep2Prob(j, in, out, aspect, thetaF, gridData);
      if rand() < prob2
        changed = true;
        aspect.Log('  Step 2: accepted');
      else
        thetaF = this.thetaFPrev;
        this.stateModel.MHSetCurrentKernels(thetaF);
        aspect.Log('  Step 2: rejected');
      end
      gramMatF = this.stateModel.MHCreateGramMatrix(j, in, out, aspect, thetaF, gridData);

      % Step 3
      f = this.DrawStateGPSurface(j, in, out, aspect, this.fPrev, thetaF, gridData);
      prob3 = this.CalcStep3Prob(j, in, out, aspect, f, thetaF, gridData);
      if rand() < prob3
        changed = true;
        aspect.Log('  Step 3: accepted');
      else
        f = this.fPrev;
        aspect.Log('  Step 3: rejected');
      end
      this.stateModel.MHSetEqnVals(f);

      % Step 4
      g = this.DrawObsGPSurface(j, in, out, aspect, this.gPrev, thetaG, gridData);
      prob4 = this.CalcStep4Prob(j, in, out, aspect, g, thetaG, gridData);
      if rand() < prob4
        changed = true;
        aspect.Log('  Step 4: accepted');
      else
        g = this.gPrev;
        aspect.Log('  Step 4: rejected');
      end
      this.obsModel.MHSetEqnVals(g);

      % Step 5

      % Nothing to do if state sequences are known.
      fSpline = this.stateModel.MHSpline(j, in, out, aspect, gridData, f);
      gSpline = this.obsModel.MHSpline(j, in, out, aspect, gridData, g);
      logLikelihood = this.likelihoodCalculator.LogLikelihood(fSpline, gSpline);
      if (isa(this.likelihoodCalculator, 'ParticleFilter') && this.likelihoodCalculator.OGC(end) > 0)
        aspect.Log('  Total %d particles are out of the grid', this.likelihoodCalculator.OGC(end));
      end
      aspect.Log('  Log likelihood = %f', logLikelihood);

      prob5 = 1;

      %%%

      accepted = changed;

      if j == 1
        message = '';
      elseif accepted
        message = ', accepted';
      else
        message = ', rejected';
      end
      aspect.Log('  Acceptance probabilities = [%f, %f, %f, %f, %f]%s', ...
        prob1, prob2, prob3, prob4, prob5, message);

      if accepted
        this.stateModel.OnAccept(j, in, out, aspect);
        this.obsModel.OnAccept(j, in, out, aspect);
        this.likelihoodCalculator.OnAccept(j, in, out, aspect);
        out.AddFnState(f, j);
        out.AddFnObs(g, j);
        this.mcdcStrategy.SaveGramMatrices(gramMatF, gramMatG);
      end

      out.SetIDXAsLast(j);
      out.j = j;

      aspect.EndIteration();

    end

  end

  methods (Access = private)

    function kernels = BuildKernelFromParams(this, params)
      params = permute(params, [2 3 1]);
      for i = 1:size(params, 1)
        kernels{i} = RBFKernel(params(i, 1), params(i, 2));
      end
    end

    function prob = CalcStep1Prob(this, j, in, out, aspect, kernels, gridData)
      initialMean = this.mcdcStrategy.InitialObsMean();
      gramMatrix = this.obsModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);
      prevGramMatrix = this.obsModel.MHCreateGramMatrix(j, in, out, aspect, this.thetaGPrev, gridData);
      numer = this.algorithm.CalcLogProbability( ...
        this.gPrev, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      denom = this.algorithm.CalcLogProbability( ...
        this.gPrev, gridData, in.gridDimForGramMatrix, initialMean, prevGramMatrix);
      aspect.Log('  Step1: numer = %f, denom = %f', numer, denom);
      prob = exp(numer - denom);
    end

    function prob = CalcStep2Prob(this, j, in, out, aspect, kernels, gridData)
      initialMean = this.mcdcStrategy.InitialStateMean();
      gramMatrix = this.stateModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);
      prevGramMatrix = this.stateModel.MHCreateGramMatrix(j, in, out, aspect, this.thetaFPrev, gridData);
      numer = this.algorithm.CalcLogProbability( ...
        this.fPrev, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      denom = this.algorithm.CalcLogProbability( ...
        this.fPrev, gridData, in.gridDimForGramMatrix, initialMean, prevGramMatrix);
      aspect.Log('  Step2: numer = %f, denom = %f', numer, denom);
      prob = exp(numer - denom);
    end

    function surface = DrawStateGPSurface(this, j, in, out, aspect, previous, kernels, gridData)
      gramMatrix = this.stateModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);
      surface = this.algorithm.DrawGPSurface( ...
        gridData, ...
        in.gridDimForGramMatrix, ...
        previous, ...
        gramMatrix);
    end

    function surface = DrawObsGPSurface(this, j, in, out, aspect, previous, kernels, gridData)
      gramMatrix = this.obsModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);
      surface = this.algorithm.DrawGPSurface( ...
        gridData, ...
        in.gridDimForGramMatrix, ...
        previous, ...
        gramMatrix);
    end

    function prob = CalcStep3Prob(this, j, in, out, aspect, f, kernels, gridData)
      initialMean = this.mcdcStrategy.InitialStateMean();
      gramMatrix = this.stateModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);

      fSpline = this.stateModel.MHSpline(j, in, out, aspect, gridData, f);
      fPrevSpline = this.stateModel.MHSpline(j, in, out, aspect, gridData, this.fPrev);

if aspect.Debug
      fig = figure;
      xs = gridData.Grids{1}';
      hold on;
      plot(xs, fSpline(xs), 'b-');
      plot(xs, fPrevSpline(xs), 'k-');
      plot(xs, mod(xs + 0.1, 2 * pi), 'r-');
      ylim([-40 40]);
      title(sprintf('Proposed f in iteration %d', j));
      xlabel('State');
      ylabel('next State');
      saveas(fig, sprintf('images/f_%04d.png', j));
      hold off;
      close(fig);
end

      c1numer = this.likelihoodCalculator.MHStateLogLikelihood(fSpline);
      c1denom = this.likelihoodCalculator.MHStateLogLikelihood(fPrevSpline);
      c2numer = this.algorithm.CalcLogProbability( ...
        f, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      c2denom = this.algorithm.CalcLogProbability( ...
        this.fPrev, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      aspect.Log('  Step3.C1: numer = %f, denom = %f', c1numer, c1denom);
      aspect.Log('  Step3.C2: numer = %f, denom = %f', c2numer, c2denom);
      prob = exp(c1numer - c1denom + c2numer - c2denom);
    end

    function prob = CalcStep4Prob(this, j, in, out, aspect, g, kernels, gridData)
      initialMean = this.mcdcStrategy.InitialObsMean();
      gramMatrix = this.obsModel.MHCreateGramMatrix(j, in, out, aspect, kernels, gridData);

      gSpline = this.obsModel.MHSpline(j, in, out, aspect, gridData, g);
      gPrevSpline = this.obsModel.MHSpline(j, in, out, aspect, gridData, this.gPrev);

if aspect.Debug
      fig = figure;
      xs = gridData.Grids{1}';
      hold on;
      plot(xs, gSpline(xs), 'b-');
      plot(xs, gPrevSpline(xs), 'k-');
      plot(xs, sin(xs), 'r-');
      ylim([-5 5]);
      title(sprintf('Proposed g in iteration %d', j));
      xlabel('State');
      ylabel('Observation');
      saveas(fig, sprintf('images/g_%04d.png', j));
      hold off;
      close(fig);
end

      c1numer = this.likelihoodCalculator.MHObsLogLikelihood(gSpline);
      c1denom = this.likelihoodCalculator.MHObsLogLikelihood(gPrevSpline);
      c2numer = this.algorithm.CalcLogProbability( ...
        g, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      c2denom = this.algorithm.CalcLogProbability( ...
        this.gPrev, gridData, in.gridDimForGramMatrix, initialMean, gramMatrix);
      aspect.Log('  Step4.C1: numer = %f, denom = %f', c1numer, c1denom);
      aspect.Log('  Step4.C2: numer = %f, denom = %f', c2numer, c2denom);
      prob = exp(c1numer - c1denom + c2numer - c2denom);
    end

  end

end
