%%% Monte Carlo Dynamic Classifier tool - MCDC tool
% http://www.ismvideo.org/NDE0UP/MCDCtool/en/index.html

% Copyright (C) 2014-2015 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, see <http://www.gnu.org/licenses/>.

%
% KnownSequence.m
%
classdef KnownSequence < LikelihoodCalculator

  properties (Access = private)
    x
    xaux
    u
    y
    q
    r
    K
    currentLogLikelihood
  end

  methods (Access = public)

    function this = KnownSequence(x, xaux, u, y, q, r, K)
      this.x    = x;
      this.xaux = xaux;
      if isempty(u)
        this.u  = zeros(size(x));
      else
        this.u  = u;
      end
      this.y    = y;
      this.q    = q;
      this.r    = r;
      this.K    = K;
    end

    function logLikelihood = LogLikelihood( ...
        this, ...
        fnState, ...
        fnObs, ...
        stateTransitionNoise, ...
        observationNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      if nargin < 4
        stateTransitionNoise = NormalDistribution(0, this.q, size(this.x, 2));
        observationNoise = NormalDistribution(0, this.r, size(this.y, 2));
      end
      xIn      = [this.x this.xaux];
      xRes     = this.x(2:end, :) - fnState(xIn(1:end-1, :)) - this.u(2:end, :);
      xLogProb = stateTransitionNoise.LogPdf(xRes); % xLogProb: T-1 double vector
      yRes     = this.y(this.K+1:end, :) - fnObs(xIn(1:end-this.K, :));
      yLogProb = observationNoise.LogPdf(yRes); % yLogProb: T double vector

      % We record logLikelihood to the property of this object.
      % It's just for logging purpose (see OnAccept below.)
      this.currentLogLikelihood = sum(xLogProb) + sum(yLogProb);
      logLikelihood = this.currentLogLikelihood;
    end

    function OnAccept(this, j, in, out, aspect)
      out.AddXE(this.x);
      out.AddYE(this.y);
      out.AddLoglik(this.currentLogLikelihood);
    end

    function logLikelihood = MHStateLogLikelihood( ...
        this, ...
        fnState, ...
        stateTransitionNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      if nargin < 3
        stateTransitionNoise = NormalDistribution(0, this.q, size(this.x, 2));
      end
      xIn      = [this.x this.xaux];
      xRes     = this.x(2:end, :) - fnState(xIn(1:end-1, :)) - this.u(2:end, :);
      xLogProb = stateTransitionNoise.LogPdf(xRes); % xLogProb: T-1 double vector

      % We record logLikelihood to the property of this object.
      % It's just for logging purpose (see OnAccept below.)
      this.currentLogLikelihood = sum(xLogProb);
      logLikelihood = this.currentLogLikelihood;
    end

    function logLikelihood = MHObsLogLikelihood( ...
        this, ...
        fnObs, ...
        observationNoise)

      % Calculate the joint probability of the sequence x(1:T) and y(1:T)
      % under the state space model:
      %   x(t) = fnState([x(t-1) xaux(t-1)]) + u(t) + q
      %   y(t+K) = fnObs(x(t)) + r
      %   q ~ stateTransitionNoise
      %   r ~ observationNoise
      % Notes
      %  1. for q and r, normal distribution is used unless otherwise specified.
      %  2. K is introduced as 'observation offset,' but we use K=0 usually.

      if nargin < 3
        observationNoise = NormalDistribution(0, this.r, size(this.y, 2));
      end
      xIn      = [this.x this.xaux];
      yRes     = this.y(this.K+1:end, :) - fnObs(xIn(1:end-this.K, :));
      yLogProb = observationNoise.LogPdf(yRes); % yLogProb: T double vector

      % We record logLikelihood to the property of this object.
      % It's just for logging purpose (see OnAccept below.)
      this.currentLogLikelihood = sum(yLogProb);
      logLikelihood = this.currentLogLikelihood;
    end

  end

end
